# -*- coding: utf-8 -*-

"""
    Copyright 2010 cloudControl UG (haftungsbeschraenkt)

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from datetime import datetime
import sys
import json
import pipes

if sys.version_info < (2, 6):
    has_str_format = False
else:
    has_str_format = True


def print_list_apps(apps):
    """
        Print a nice table of apps.
    """
    print 'Apps'
    if has_str_format:
        print " {0:3} {1:30} {2:6}".format('Nr', 'Name', 'Type')
        for i, app in enumerate(apps):
            print " {0:3} {1:30} {2:6}".format(
                i + 1,
                app['name'],
                app['type']['name'])
    else:
        print " %-3ls %-30ls %-6ls" % ('Nr', 'Name', 'Type')
        for i, app in enumerate(apps):
            print " %-3ls %-30ls %-6ls" % (
                i + 1,
                app['name'],
                app['type']['name'])


def print_app_details(app, public_key):
    """
        Print app details.
    """
    print 'App'
    if has_str_format:
        print " Name: {0:30} Type: {1:10} Owner: {2:20}".format(
            app['name'],
            app['type']['name'],
            app['owner']['username'])
        if app['type']['name'] == 'custom':
            print ' Buildpack URL: {0}\n'.format(app['buildpack_url'])
        print ' Repository: {0}'.format(app['repository'])
        print '\n Public Key: {0}'.format(public_key)

        print '\n Users'
        print "   {0:15} {1:35} {2:10} {3:10}".format('Name', 'Email', 'Role', 'Deployment')
        for user in app['users']:
            print "   {0:15} {1:35} {2:10} {3:10}".format(
                user['username'],
                user['email'],
                user['role'] if 'role' in user else '',
                user['deployment'] if 'deployment' in user else '(app)'
            )
        print '\n Deployments'
        for deployment in app['deployments']:
            print "   {0:60}".format(deployment['name'])
    else:
        print " Name: %-30ls Type: %-10ls Owner: %-20ls" % (
            app['name'],
            app['type']['name'],
            app['owner']['username'])
        print ' Repository: %s' % (app['repository'])
        print '\n Public Key: %s' % public_key
        print '\n Users'
        print "   %-15ls %-35ls %-10ls %-10ls" % ('Name', 'Email', 'Role', 'Deployment')
        for user in app['users']:
            if 'deployment' not in user:
                user['deployment'] = '(app)'
            if 'role' not in user:
                user['role'] = ''

            print "   %(username)-15ls %(email)-35ls %(role)-10ls %(deployment)-10ls" % (user)
        print '\n Deployments'
        for deployment in app['deployments']:
            print "   %(name)-60ls" % (deployment)


def print_deployment_details(deployment):
    """
        Print deployment details.
    """
    app, dep = deployment['name'].split("/")
    url = "http://{0}".format(deployment['default_subdomain'])

    if deployment['stack']:
        stack = deployment['stack']['name']
    else:
        stack = None
    print 'Deployment'
    if has_str_format:
        print ' name: {0}'.format(deployment['name'])
        print ' stack: {0}'.format(stack)
        print ' URL: {0}'.format(url)
        print ' branch: {0}'.format(deployment['branch'])
        print ' last modified: {0}'.format(deployment['date_modified'])
        print ' current version: {0}'.format(deployment['version'])
        print ' current state: {0}'.format(deployment['state'])
        print ' containers: {0}'.format(deployment['min_boxes'])
        print ' memory: {0}MB'.format(deployment['max_boxes'] * 128)

        if 'users' in deployment and deployment['users']:
            print '\n Users'
            print "   {0:15} {1:35} {2:10} {3:10}".format('Name', 'Email', 'Role', 'Deployment')
            for user in deployment['users']:
                print "   {0:15} {1:35} {2:10} {3:10}".format(
                    user['username'],
                    user['email'],
                    user['role'],
                    '(app)' if 'app' in user else deployment['name'],
                )

    else:
        print ' name: %(name)s' % (deployment)
        print ' stack: %s' % (stack)
        print ' URL: %s' % (url)
        print ' branch: %(branch)s' % (deployment)
        print ' last modified: %(date_modified)s' % (deployment)
        print ' current version: %(version)s' % (deployment)
        print ' current state: %(state)s' % (deployment)
        print ' containers: %(min_boxes)s' % (deployment)
        print ' memory: %(min_boxes)sMB' % (deployment) * 128

        if 'users' in deployment and deployment['users']:
            print '\n Users'
            print "   %-15ls %-35ls %-10ls %-10ls" % ('Name', 'Email', 'Role', 'Deployment')
            for user in deployment['users']:
                user['deployment'] = '(app)' if 'app' in user else deployment['name']
                print "   %(username)-15ls %(email)-35ls %(role)-10ls %(deployment)-10ls" % (user)


def print_user_list_app(app):
    print 'Users'
    if has_str_format:
        print "   {0:15} {1:35} {2:10} {3:10}".format('Name', 'Email', 'Role', 'Deployment')
        for user in app['users']:
            print "   {0:15} {1:35} {2:10} {3:10}".format(
                user['username'],
                user['email'],
                user['role'] if 'role' in user else '',
                user['deployment'] if 'deployment' in user else '(app)'
            )
    else:
        print "   %-15ls %-35ls %-10ls %-10ls" % ('Name', 'Email', 'Role', 'Deployment')
        for user in app['users']:
            if 'deployment' not in user:
                user['deployment'] = '(app)'
            if 'role' not in user:
                user['role'] = ''

            print "   %(username)-15ls %(email)-35ls %(role)-10ls %(deployment)-10ls" % (user)


def print_user_list_deployment(deployment):
    print 'Users'
    if has_str_format:
        print "   {0:15} {1:35} {2:10} {3:10}".format('Name', 'Email', 'Role', 'Deployment')
        for user in deployment['users']:
            print "   {0:15} {1:35} {2:10} {3:10}".format(
                user['username'],
                user['email'],
                user['role'],
                '(app)' if 'app' in user else deployment['name'],
            )
    else:
        print '\n Users'
        print "   %-15ls %-35ls %-10ls %-10ls" % ('Name', 'Email', 'Role', 'Deployment')
        for user in deployment['users']:
            user['deployment'] = '(app)' if 'app' in user else deployment['name']
            print "   %(username)-15ls %(email)-35ls %(role)-10ls %(deployment)-10ls" % (user)


def print_alias_list(aliases):
    """
        Print a list of aliases
    """
    if has_str_format:
        print 'Aliases'
        print ' {0:60} {1:8} {2:8}'.format('name', 'default', 'verified')
        for alias in aliases:
            print ' {0:60} {1:8} {2:8}'.format(
                alias['name'],
                alias['is_default'],
                alias['is_verified'])
    else:
        print 'Aliases'
        print ' %-60ls %-8ls %-8ls' % ('name', 'default', 'verified')
        for alias in aliases:
            print ' %(name)-60ls %(is_default)-8ls %(is_verified)-8ls' % (
                alias)


def print_alias_details(alias):
    """
        Print alias details.
    """
    if has_str_format:
        print '{0:28}: {1}'.format('Alias', alias['name'])
        print '   {0:25}: {1}'.format(
            'is_default',
            str(alias['is_default']))
        print '   {0:25}: {1}'.format(
            'is_verified',
            str(alias['is_verified']))
        print '   {0:25}: {1}'.format(
            'verification_errors',
            alias['verification_errors'])
        print '   {0:25}: {1}'.format(
            'verification_code',
            alias['verification_code'])
        print '   {0:25}: {1}'.format(
            'date_created',
            alias['date_created'])
        print '   {0:25}: {1}'.format(
            'date_modified',
            alias['date_modified'])
    else:
        print '%-28ls: %s' % (
            'Alias',
            alias['name'])
        print '   %-25ls: %s' % (
            'is_default',
            str(alias['is_default']))
        print '   %-25ls: %s' % (
            'is_verified',
            str(alias['is_verified']))
        print '   %-25ls: %s' % (
            'verification_errors',
            alias['verification_errors'])
        print '   %-25ls: %s' % (
            'verification_code',
            alias['verification_code'])
        print '   %-25ls: %s' % (
            'date_created',
            alias['date_created'])
        print '   %-25ls: %s' % (
            'date_modified',
            alias['date_modified'])


def print_log_entries(logEntries, apache_type):
    """
        Print log lines that fit the apache type.

        Either access or error.
    """
    if apache_type == 'access':
        if has_str_format:
            for entry in logEntries:
                dt = datetime.fromtimestamp(float(entry["time"]))
                entry["time"] = dt.strftime('[%d/%b/%Y:%H:%M:%S +0000]')
                try:
                    print r'{0} {1} {2} {3} "{4}" {5} {6} "{7}" "{8}"'.format(
                        entry["remote_host"],
                        entry["remote_logname"],
                        entry["remote_user"],
                        entry["time"],
                        entry["first_request_line"],
                        entry["status"],
                        entry["response_size_CLF"],
                        entry["referer"],
                        entry["user_agent"])
                except KeyError:
                    pass
        else:
            for entry in logEntries:
                dt = datetime.fromtimestamp(float(entry["time"]))
                entry["time"] = dt.strftime('[%d/%b/%Y:%H:%M:%S +0000]')
                try:
                    print r'%s %s %s %s "%s" %s %s "%s" "%s"' % (
                        entry["remote_host"],
                        entry["remote_logname"],
                        entry["remote_user"],
                        entry["time"],
                        entry["first_request_line"],
                        entry["status"],
                        entry["response_size_CLF"],
                        entry["referer"],
                        entry["user_agent"])
                except KeyError:
                    pass
    elif apache_type == 'error':
        if has_str_format:
            for entry in logEntries:
                dt = datetime.fromtimestamp(float(entry["time"]))
                entry["time"] = dt.strftime('[%a %b %d %H:%M:%S %Y]')
                print r'{0} {1} {2}'.format(
                    entry["time"],
                    entry["type"],
                    entry["message"].encode('utf-8'))
        else:
            for entry in logEntries:
                dt = datetime.fromtimestamp(float(entry["time"]))
                entry["time"] = dt.strftime('[%a %b %d %H:%M:%S %Y]')
                print r'%s %s %s' % (
                    entry["time"],
                    entry["type"],
                    entry["message"].encode('utf-8'))
    elif apache_type == 'worker':
        if has_str_format:
            for entry in logEntries:
                dt = datetime.fromtimestamp(float(entry["time"]))
                entry["time"] = dt.strftime('[%a %b %d %H:%M:%S %Y]')
                print r'{0} {1} {2}'.format(
                    entry["time"],
                    entry["wrk_id"],
                    entry["message"].encode('utf-8'))
        else:
            for entry in logEntries:
                dt = datetime.fromtimestamp(float(entry["time"]))
                entry["time"] = dt.strftime('[%a %b %d %H:%M:%S %Y]')
                print r'%s %s %s' % (
                    entry["time"],
                    entry["wrk_id"],
                    entry["message"].encode('utf-8'))
    elif apache_type == 'deploy':
        if has_str_format:
            for entry in logEntries:
                dt = datetime.fromtimestamp(float(entry["time"]))
                entry["time"] = dt.strftime('[%a %b %d %H:%M:%S %Y]')
                print r'{0} {1} {2} {3}'.format(
                    entry["time"],
                    entry['hostname'],
                    entry["level"],
                    entry["message"].encode('utf-8'))
        else:
            for entry in logEntries:
                dt = datetime.fromtimestamp(float(entry["time"]))
                entry["time"] = dt.strftime('[%a %b %d %H:%M:%S %Y]')
                print r'%s %s %s %s' % (
                    entry["time"],
                    entry['hostname'],
                    entry["level"],
                    entry["message"].encode('utf-8'))


def print_keys(keys):
    """
        Print a list of keys.
    """
    print 'Keys'
    if has_str_format:
        for key in keys:
            print ' {0:15}'.format(key['key_id'])
    else:
        for key in keys:
            print ' %(key_id)-15ls' % (key)


def print_key(key):
    """
        Print a users public key.
    """
    print key["key"]


def print_addons(addons):
    """
       Print all available addons.
    """
    if has_str_format:
        for addon in addons:
            print '\nAddon: {0}'.format(addon['name'])
            for option in addon['options']:
                print '   {0}'.format(option['name'])
    else:
        for addon in addons:
            print '\nAddon: %s' % (addon['name'])
            for option in addon['options']:
                print '   %s' % option['name']


def print_addon_creds(addons):
    creds_format = {}
    for addon in addons:
        if len(addon['settings']):
            name = addon['addon_option']['name'].split('.')[0].upper()
            creds_format[name] = addon['settings']
    print json.dumps(creds_format, indent=4, sort_keys=True)


def print_addon_list(addons):
    """
        Print a list of addon details
    """
    for addon in addons:
        print_addon_details(addon)
        print '\n'


def print_addon_details(addon):
    """
        Print addon details.
    """
    if has_str_format:
        print '{0:25}: {1}'.format('Addon', addon['addon_option']['name'])
        if len(addon['settings']) > 0:
            print '   \n Settings'
            for key, value in addon['settings'].items():
                print '   {0:25}: {1}'.format(key, value)
    else:
        print '%-25ls: %s' % ('Addon', addon['addon_option']['name'])
        if len(addon['settings']) > 0:
            print '   \n Settings'
            for key, value in addon['settings'].items():
                print '   %-25ls: %s' % (key, value)


def print_worker_list(workers):
    print 'Workers'
    if has_str_format:
        print ' {0:3} {1:11} {2}'.format('nr.', 'wrk_id', 'command')
        for count, worker in enumerate(workers):
            print ' {0:3} {1:11} {2}'.format(count + 1, worker['wrk_id'], worker['command'])
    else:
        print ' %-3ls %-11ls %s' % ('nr.', 'wrk_id', 'command')
        for count, worker in enumerate(workers):
            print ' %-3ls %-11ls %s' % (count + 1, worker['wrk_id'], worker['command'])


def print_worker_details(worker):
    print 'Worker'
    if has_str_format:
        print ' {0:9}: {1}'.format('wrk_id', worker['wrk_id'])
        try:
            print ' {0:9}: {1}'.format('command', worker['command'].encode('utf-8'))
        except UnicodeDecodeError:
            print ' {0:9}: {1}'.format('command', worker['command'])
        print ' {0:9}: {1}'.format('params', worker['params'])
        print ' {0:9}: {1}'.format('size', worker['size'])
        print ' {0:9}: {1}'.format('created', worker['date_created'])
    else:
        print ' %-9ls: %s' % ('wrk_id', worker['wrk_id'])
        print ' %-9ls: %s' % ('command', worker['command'])
        print ' %-9ls: %s' % ('params', worker['params'])
        print ' %-9ls: %s' % ('size', worker['size'])
        print ' %-9ls: %s' % ('created', worker['date_created'])


def print_cronjob_list(cronjobs):
    print 'Cronjobs'
    if has_str_format:
        print ' {0:3} {1:11}'.format('nr.', 'job_id')
        for count, cronjob in enumerate(cronjobs):
            print ' {0:3} {1:11}'.format(count + 1, cronjob['job_id'])
    else:
        print ' %-3ls %-11ls' % ('nr.', 'job_id')
        for count, cronjob in enumerate(cronjobs):
            job_id = cronjob['wrk_id']
            print ' %-3ls %-11ls' % (count + 1, job_id)


def print_cronjob_details(cronjob):
    print 'Cronjob'
    if has_str_format:
        print ' {0:9}: {1}'.format('job_id', cronjob['job_id'])
        print ' {0:9}: {1}'.format('url', cronjob['url'])
        print ' {0:9}: {1}'.format('next_run', cronjob['next_run'])
        print ' {0:9}: {1}'.format('created', cronjob['date_created'])
        print ' {0:9}: {1}'.format('modified', cronjob['date_modified'])
    else:
        print ' %-9ls: %s' % ('job_id', cronjob['job_id'])
        print ' %-9ls: %s' % ('url', cronjob['url'])
        print ' %-9ls: %s' % ('next_run', cronjob['next_run'])
        print ' %-9ls: %s' % ('created', cronjob['date_created'])
        print ' %-9ls: %s' % ('modified', cronjob['date_modified'])


def get_version(cctrlversion, cclibversion):
    """
        Prepare the version string
    """
    if has_str_format:
        return '%(prog)s {0} using pycclib {1}'.format(
            cctrlversion,
            cclibversion)
    else:
        return '%%(prog)s %s using pycclib %s' % (cctrlversion, cclibversion)


def print_config(config, key=None):
    if key and key in config:
        print config.get(key)
    elif key:
        print '[ERROR] Key `{}` not found.'.format(key)
    else:
        for key in sorted(config):
            print u'{0}={1}'.format(key, pipes.quote(config.get(key) or ''))



# -*- coding: utf-8 -*-
"""
    Copyright 2010 cloudControl UG (haftungsbeschraenkt)

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
import sys

messages = {}
messages['WrongApplication'] = r'This application is unknown.'
messages['WrongDeployment'] = r'This deployment is unknown.'
messages['PasswordsDontMatch'] = r"The passwords don't match."
messages['InvalidApplicationName'] = r'Name may only contain a-z and 0-9 and must not start with a number.'
messages['WrongUsername'] = r'This username is unknown.'
messages['UserBelongsToApp'] = r'This user already belongs to this application.'
messages['RemoveUserGoneError'] = r'No such app or user. Please check app name and user name or email address.'
messages['UserCreatedNowCheckEmail'] = r'User has been created. Please check you e-mail for your confirmation code.'
messages['DeleteOnlyApplication'] = r'You can only delete applications not deployments. Try the undeploy command.'
messages['NoAliasGiven'] = r'You have to specify an alias.'
messages['WrongAlias'] = r'This alias is unknown.'
messages['NotAllowed'] = r'Sorry. You are not allowed to perform this action.'
messages['CannotDeleteDeploymentExist'] = r'You have to undeploy all related deployments, before you can delete the application.'
messages['NotAuthorized'] = r'The authorization failed, check your e-mail address and password.'
messages['PermissionDenied'] = r'You are not allowed to push to this repository. Maybe check your keys using "cctrluser key".'
messages['SecurityQuestionDenied'] = r'Action canceled on user request.'
messages['WrongAddon'] = r'This addon is unknown for this app_name/deployment_name.'
messages['DuplicateAddon'] = r'You can not add the same addon option twice.'
messages['InvalidAddon'] = r'This is not a valid addon name. Check the list of available addons with {} app_name/deployment_name addon.list'.format(os.path.basename(sys.argv[0]))
messages['ForbiddenAddon'] = 'You are not allowed to perform this action.\nIf you are trying to use a Beta addon, you can request access from the addon page.'
messages['WrongPubKeyName'] = r'The public key file must be named "id_rsa.pub".'
messages['NoSuchKeyFile'] = r'No such key file. Please check your input.'
messages['WrongKeyFormat'] = r'Your id_rsa.pub public key file seems to be in the wrong format.'
messages['InvalidAppOrDeploymentName'] = r'The application or deployment name is invalid.'
messages['KeyDuplicate'] = r'This key was added previously.'
messages['NoWorkerCommandGiven'] = r'The worker command is missing. Try the path to your PHP file relative from your repository root.'
messages['NoRunCommandGiven'] = r'Run command is missing.'
messages['WrongWorker'] = r'There is no such worker for this app_name/deployment_name.'
messages['NeitherBazaarNorGitFound'] = r'Cannot find "git" nor "bzr"! Please make sure either Bazaar or Git executables are in your path.'
messages['BazaarRequiredToPush'] = r'Please make sure the Bazaar executable is in your path.'
messages['GitRequiredToPush'] = r'Please make sure the Git executable is in your path.'
messages['NoCronURLGiven'] = r'You must provide a URL for cron to call.'
messages['NoSuchCronJob'] = r'Sorry, we can not find cronjob with this ID.'
messages['FileReadOrWriteFailed'] = r'Sorry, could not read or write to file.'
messages['FileNotFound'] = r'Sorry, file not found!'
messages['UserShouldCreateKey'] = r'Sorry, something went wrong when creating a key. Please create a key on your system, then run the command again.'
messages['BazaarConfigFound'] = r'Bazaar configuration found! Using "Bazaar" as repository type.'
messages['GitConfigFound'] = r'Git configuration found! Using "Git" as repository type.'
messages['BazaarExecutableFound'] = r'Bazaar seems to be installed! Using "Bazaar" as repository type.'
messages['GitExecutableFound'] = r'Git seems to be installed! Using "Git" as repository type.'
messages['CreatingAppAsDefaultRepoType'] = r'Using default "Git" as repository type.'
messages['DeleteAppsBeforeUser'] = r'There are still applications associated with this user account. Undeploy and/or delete applications before deleting user.'
messages['NoSuchFile'] = r'File not found.'
messages['APIUnreachable'] = r'Could not connect to API...'
messages['NoBuildpackURL'] = r'You need to provide a buildpack URL for "custom" application type'
messages['NoCustomApp'] = r'You can only provide a buildpack URL if the app type is "custom"'
messages['NoValidBuildpackURL'] = r'The buildpack URL provided is not valid. Please try again.'
messages['AmbiguousSize'] = r'You can only specify one of --size or --memory'
messages['InvalidMemory'] = r'Memory size should be an integer between 128 and 1024 MB'
messages['InvalidSize'] = r'Size should be an integer between 1 and 8'
messages['NoPreviousVersionFound'] = r'Previous version not found.'
messages['ClearCacheFailed'] = r'Clear buildpack cache failed.'
messages['DeploymentFailed'] = r'Deployment failed.'
messages['CommandNotImplemented'] = r'Sorry, this command is not available.'
messages['ShipAndDeploy'] = r'--ship and --push options cannot be used simultaneously.'
messages['RegisterDisabled'] = r'You can register on {}.'
messages['NoVariablesGiven'] = r'You must provide some variables.'
messages['DuplicatedFlag'] = r'Please, specify a flag only once.'

if sys.platform == 'win32':
    messages['UpdateAvailable'] = r'A newer version is available. Please update.'
    messages['UpdateRequired'] = r'A newer version is required. You need to upgrade before using this program.'
else:
    messages['UpdateAvailable'] = r'A newer version is available. To upgrade run: (sudo) pip install {} --upgrade'
    messages['UpdateRequired'] = r'A newer version is required. You need to upgrade before using this program. To upgrade run: (sudo) pip install {} --upgrade'


class InputErrorException(Exception):
    """
        This exception is raised if for some reason someone put something in we
        could not understand at all.
    """
    def __init__(self, errorKey):
        self.error_message = messages[errorKey]

    def __str__(self):
        return '[ERROR]' + ' ' + self.error_message


class PasswordsDontMatchException(Exception):
    """
        This exception is raised if the password and the password check weren't
        equal for three times.
    """
    pass



# -*- coding: utf-8 -*-
"""
    Copyright 2010 cloudControl UG (haftungsbeschraenkt)

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __builtin__ import raw_input, str, filter, float, len
import exceptions
import os
import time
import re
import subprocess
import shlex
import webbrowser
import math
import sys
import urlparse

from settings import CONFIG_ADDON
from datetime import datetime, timedelta

from pycclib.cclib import GoneError, ForbiddenError, TokenRequiredError, \
    BadRequestError, ConflictDuplicateError, UnauthorizedError, \
    NotImplementedError, ThrottledError
from subprocess import check_call, CalledProcessError
from cctrl.error import InputErrorException, messages
from cctrl.oshelpers import check_installed_rcs, is_buildpack_url_valid, \
    ssh_cmd, check_output
from cctrl.output import print_deployment_details, print_app_details,\
    print_alias_details, print_log_entries, print_list_apps,\
    print_addon_details, print_addons, print_addon_list, print_alias_list, \
    print_worker_list, print_worker_details, print_cronjob_list, \
    print_cronjob_details, print_addon_creds, print_config
from output import print_user_list_app, print_user_list_deployment
from cctrl.addonoptionhelpers import parse_additional_addon_options, \
    parse_config_variables, extract_flag_from_variables


class AppsController():
    """
        This controller handles the special case where you want to get a
        list of applications.
    """

    api = None

    def __init__(self, api):
        self.api = api

    def list(self):  # @ReservedAssignment
        apps = self.api.read_apps()
        print_list_apps(apps)


class CVSType():
    """
        A simple class for both supported repository types.
    """
    GIT = 'git'
    BZR = 'bzr'

    @staticmethod
    def by_path(application_path):
        """
            Provides the cvs (repo) type by checking if given directory
            contains a ".git" or ".bzr" configuration directory.
        """
        for (dirname, cvstype, msg) in [(".git", CVSType.GIT, 'GitConfigFound'),
                                        (".bzr", CVSType.BZR, 'BazaarConfigFound')]:
            if os.path.exists(os.path.join(application_path, dirname)):
                return (cvstype, msg)

        return (None, None)

    @staticmethod
    def by_env():
        """
            Provides the cvs (repo) type by checking environment variable
            PATH for existence of either Bazaar or Git.
        """
        for (execname, cvstype, msg) in [('git', CVSType.GIT, 'GitExecutableFound'),
                                         ('bzr', CVSType.BZR, 'BazaarExecutableFound')]:
            if check_installed_rcs(execname):
                return (cvstype, msg)

        return (None, None)


class AppController():
    """
        After parsing the command line in parse_cmdline() the related
        method of the ApplicationController gets called. Each method uses
        pycclib to fire a request and handles the response showing it to the
        user if needed.
    """
    def __init__(self, api, settings):
        self.api = api
        self.settings = settings

    def run_cmd(self, args):
        try:
            app_name, deployment_name = self.parse_app_deployment_name(args.name)
        except ParseAppDeploymentName:
            raise InputErrorException('InvalidApplicationName')

        if not deployment_name:
            deployment_name = 'default'

        user_host = '{app}-{dep}@{host}'.format(app=app_name, dep=deployment_name, host=self.settings.ssh_forwarder)

        # Refresh our token before sending it to the forwarder.
        try:
            self.api.read_deployment(app_name, deployment_name)
        except GoneError:
            raise InputErrorException('WrongApplication')
        env = 'TOKEN={token}'.format(token=self.api.get_token()['token'])
        if len(args.command) > 0:
            command = '{env} {command}'.format(env=env, command=args.command)
        else:
            raise InputErrorException('NoRunCommandGiven')
        sshopts = shlex.split(os.environ.get('CCTRL_SSHOPTS', ''))
        ssh_command = ['ssh', '-t'] + sshopts + ['-p', self.settings.ssh_forwarder_port, '--', user_host, command]
        subprocess.call(ssh_command)

    def rollback_cmd(self, args):
        """
           Rollback command
        """

        try:
            app_name, deployment_name = self.parse_app_deployment_name(args.name)
        except ParseAppDeploymentName:
            raise InputErrorException('InvalidApplicationName')

        if not deployment_name:
            deployment_name = 'default'

        logEntries = []

        try:
            logEntries = self.api.read_log(
                app_name,
                deployment_name,
                'deploy',
                last_time=None)
        except GoneError:
            raise InputErrorException('WrongApplication')

        deployments = [e['message'].split(':')[1].strip() for e in logEntries
                       if 'Deployed version' in e['message']]

        try:
            current_deployment = deployments.pop()
        except IndexError:
            raise InputErrorException('NoPreviousVersionFound')

        previous_deployment = None

        while not previous_deployment:
            try:
                deployment_version = deployments.pop()
                if deployment_version != current_deployment:
                    previous_deployment = deployment_version
            except IndexError:
                raise InputErrorException('NoPreviousVersionFound')

        try:
            self.api.update_deployment(
                app_name,
                version=previous_deployment,
                deployment_name=deployment_name)
        except GoneError:
            raise InputErrorException('WrongApplication')
        except ForbiddenError:
            raise InputErrorException('NotAllowed')

    def create(self, args):
        """
            Creates a new application.
        """
        try:
            app_name, deployment_name = self.parse_app_deployment_name(args.name)
        except ParseAppDeploymentName:
            raise InputErrorException('InvalidApplicationName')

        if args.buildpack:
            # Did the user choose a default app type and provided a buildpack url?
            if not args.type == 'custom':
                raise InputErrorException('NoCustomApp')
            # Did the user provide a valid buildpack URL?
            elif not is_buildpack_url_valid(args.buildpack):
                raise InputErrorException('NoValidBuildpackURL')
        # Did the user provide a buildpack url if app has a custom type?
        elif args.type == 'custom':
            raise InputErrorException('NoBuildpackURL')

        # Did the user provide the repo type as argument?
        if args.repo:
            repo_type = args.repo
            detection_method = None
        else:
            # No, he/she didn't! Then check if current directory is an app and already has a CVS type ...
            (repo_type, detection_method) = CVSType.by_path(os.getcwd())

            if repo_type is None:
                # Hmm, current directory was nothing. Let's check if either 'bzr' or 'git' is installed ...
                (repo_type, detection_method) = CVSType.by_env()

            if repo_type is None:
                # Hmm, also nothing installed! Ok, we give up and set default = GIT and hope for better times ...
                detection_method = 'CreatingAppAsDefaultRepoType'
                repo_type = CVSType.GIT

        try:
            self.api.create_app(app_name, args.type, repo_type, args.buildpack)
            self.api.create_deployment(
                app_name,
                deployment_name=deployment_name)
            if detection_method:
                print messages[detection_method]
        except GoneError:
            raise InputErrorException('WrongApplication')
        except ForbiddenError:
            raise InputErrorException('NotAllowed')
        else:
            return True

    def delete(self, args):
        """
            Delete an application. If we wouldn't check the token here it could
            happen that we ask the user for confirmation and then fire the api
            request. If the token wasn't valid this would result in a
            TokenRequiredError being raised and after getting the credentials
            and creating a token this method would be called a second time.

            This would result in asking the user two times if he really wants
            to delete the app which is a rather bad user experience.
        """
        if self.api.check_token():
            app_name, deployment_name = self.parse_app_deployment_name(args.name)
            if not self.does_app_exist(app_name):
                raise InputErrorException('WrongApplication')
            if deployment_name:
                raise InputErrorException('DeleteOnlyApplication')
            if not args.force_delete:
                question = raw_input("Do you really want to delete application '{0}'? ".format(app_name) +
                                     'Type "Yes" without the quotes to delete: ')
            else:
                question = 'Yes'
            if question.lower() == 'yes':
                try:
                    self.api.delete_app(app_name)
                except ForbiddenError:
                    raise InputErrorException('NotAllowed')
                except BadRequestError:
                    raise InputErrorException('CannotDeleteDeploymentExist')
                except GoneError:
                    raise InputErrorException('WrongApplication')
            else:
                print messages['SecurityQuestionDenied']
        else:
            raise TokenRequiredError

    def _details(self, app_or_deployment_name):
        app_name, deployment_name = self.parse_app_deployment_name(app_or_deployment_name)
        if deployment_name:
            try:
                deployment = self.api.read_deployment(
                    app_name,
                    deployment_name)

                try:
                    app_users = self.api.read_app_users(app_name)
                except (UnauthorizedError, ForbiddenError, NotImplementedError):
                    # ok since possibly I am not allowed to see users at all
                    pass

                else:
                    deployment['users'] = [
                        dict(au, app=True)
                        for au in app_users
                    ] + deployment['users']

            except GoneError:
                raise InputErrorException('WrongDeployment')
            else:
                return app_name, deployment_name, deployment
        else:
            try:
                app = self.api.read_app(app_name)

                # only get deployment-users if i can see app-users
                if len(app['users']):
                    try:
                        for deployment in app['deployments']:
                            appname, depname = self.parse_app_deployment_name(deployment['name'])

                            depusers = self.api.read_deployment_users(appname, depname)

                            app['users'].extend(
                                dict(du, deployment=depname)
                                for du in depusers
                            )
                    except (NotImplementedError, BadRequestError):  # for old api-servers
                        pass

            except GoneError:
                raise InputErrorException('WrongApplication')
            else:
                return app_name, deployment_name, app

    def details(self, args):
        """
            Print application or deployment details.

            e.g.:

            'cctrlapp APP_NAME details' prints application details

            'cctrlapp APP_NAME/DEP_NAME details' prints deployment details
        """
        app_name, deployment_name, obj = self._details(args.name)
        if deployment_name:
            print_deployment_details(obj)
        else:
            public_key_content = self._get_public_key(app_name, obj)
            print_app_details(obj, public_key_content)

    def _get_deployment_url(self, obj):
        return "http://{0}".format(obj.get('default_subdomain'))

    def _get_repository_url(self, obj):
        return urlparse.urlsplit(obj.get('repository')).netloc

    def get_deployment_from_app_deployment_name(self, app_or_deployment_name):
        app_name, deployment_name = self.parse_app_deployment_name(app_or_deployment_name)
        if not deployment_name:
            deployment_name = 'default'

        try:
            deployment = self.api.read_deployment(
                app_name,
                deployment_name)

        except GoneError:
            raise InputErrorException('WrongDeployment')
        else:
            return app_name, deployment_name, deployment

    def _open(self, url):
        savout = os.dup(1)
        os.close(1)
        os.open(os.devnull, os.O_RDWR)
        try:
            webbrowser.open_new_tab(url)
        finally:
            os.dup2(savout, 1)

    def open(self, args):
        """
            Open deployment URL on the default browser.

            e.g.:

            'cctrlapp APP_NAME open' opens the default deployment's URL

            'cctrlapp APP_NAME/DEP_NAME open' opens the deployment's URL
        """
        app_name, deployment_name, obj = self.get_deployment_from_app_deployment_name(args.name)
        url = self._get_deployment_url(obj)
        self._open(url)

    def _get_size_from_memory(self, memory):
        res = re.match(r'(\d+)(.*)', memory.lower())
        if not res:
            raise InputErrorException('InvalidMemory')
        if res.group(2) in ['mb', 'm', '']:
            size = float(res.group(1)) / 128
        elif res.group(2) in ['gb', 'g']:
            size = float(res.group(1)) * 2 ** 10 / 128
        else:
            raise InputErrorException('InvalidMemory')
        final_size = int(math.ceil(size))
        if final_size != size:
            print >> sys.stderr, 'Memory size has to be a multiple of 128MB and has been rounded up to {0}MB.'.format(final_size * 128)
        return final_size

    def deploy(self, args):
        """
            Deploy a distinct version.

            Since we want to make it as easy as possible we first try to update
            the default deployment and start the newest version of that if no
            other arguments were passed at the command line.
        """
        try:
            app_name, deployment_name = self.parse_app_deployment_name(args.name)
        except ParseAppDeploymentName:
            raise InputErrorException('InvalidApplicationName')
        if not deployment_name:
            deployment_name = 'default'
        if args.size:
            size = args.size
            if args.memory:
                raise InputErrorException('AmbiguousSize')
        elif args.memory:
            memory = args.memory
            size = self._get_size_from_memory(memory)
        else:
            size = None
        try:
            try:
                self.api.update_deployment(
                    app_name,
                    version=args.version,
                    deployment_name=deployment_name,
                    min_boxes=args.containers,
                    max_boxes=size,
                    stack=args.stack)
            except GoneError:
                try:
                    self.api.create_deployment(
                        app_name,
                        deployment_name=deployment_name,
                        stack=args.stack)
                    self.api.update_deployment(
                        app_name,
                        version=args.version,
                        deployment_name=deployment_name,
                        min_boxes=args.containers,
                        max_boxes=size,
                        stack=args.stack)
                except GoneError:
                    raise InputErrorException('WrongApplication')
                except ForbiddenError:
                    raise InputErrorException('NotAllowed')
        except BadRequestError as e:
            if 'max_boxes_over_max_process_limit' in e.msgs:
                if args.memory:
                    raise InputErrorException('InvalidMemory')
                if args.size:
                    raise InputErrorException('InvalidSize')
            else:
                raise

        if args.restart_workers:
            deployment = {'state': 'unknown'}
            while deployment['state'] != 'deployed':
                time.sleep(1)
                deployment = self.api.read_deployment(app_name, deployment_name)
            self._restartWorkers(app_name, deployment_name)

        return True

    def redeploy(self, deployment):
        """
            Redeploy an existing application with same options and configuration.
        """
        try:
            app_name, deployment_name = self.parse_app_deployment_name(deployment['name'])
            self.api.update_deployment(
                app_name,
                version=-1,
                deployment_name=deployment_name,
                min_boxes=deployment['min_boxes'],
                max_boxes=deployment['max_boxes'],
                stack=deployment['stack']['name'])
        except (KeyError, GoneError, BadRequestError):
            raise InputErrorException('DeploymentFailed')

    def undeploy(self, args):
        """
            Undeploys the deployment, deletes the database and files.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not self.does_app_exist(app_name):
            raise InputErrorException('WrongApplication')
        if not args.force_delete:
            question = raw_input("Do you really want to delete deployment '{0}'? ".format(args.name) +
                                 'This will delete everything including files and the database. Type "Yes" without the quotes to delete: ')
        else:
            question = 'Yes'
        if question.lower() == 'yes':
            args.force_delete = True
            try:
                self.api.delete_deployment(app_name, deployment_name)
            except GoneError:
                raise InputErrorException('WrongDeployment')
        else:
            print messages['SecurityQuestionDenied']
        return True

    def addAlias(self, args):
        """
            Adds the given alias to the deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.alias:
            raise InputErrorException('NoAliasGiven')
        self.api.create_alias(app_name, args.alias, deployment_name)
        return True

    def showAlias(self, args):
        """
            Shows the details of an alias.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.alias:
            aliases = self.api.read_aliases(app_name, deployment_name)
            print_alias_list(aliases)
            return True
        else:
            try:
                alias = self.api.read_alias(
                    app_name,
                    args.alias,
                    deployment_name)
            except GoneError:
                raise InputErrorException('WrongAlias')
            else:
                print_alias_details(alias)
                return True

    def removeAlias(self, args):
        """
            Removes an alias form a deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.alias:
            raise InputErrorException('NoAliasGiven')
        try:
            self.api.delete_alias(app_name, args.alias, deployment_name)
        except GoneError:
            raise InputErrorException('WrongAlias')
        return True

    def addWorker(self, args):
        """
            Adds the given worker to the deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if args.size:
            size = args.size
            if args.memory:
                raise InputErrorException('AmbiguousSize')
        elif args.memory:
            memory = args.memory
            size = self._get_size_from_memory(memory)
        else:
            size = None
        if not deployment_name:
            deployment_name = 'default'
        if not args.command:
            raise InputErrorException('NoWorkerCommandGiven')
        try:
            self.api.create_worker(
                app_name,
                deployment_name,
                args.command,
                params=args.params,
                size=size)
        except BadRequestError as e:
            if 'size' in e.msgs:
                if args.memory:
                    raise InputErrorException('InvalidMemory')
                if args.size:
                    raise InputErrorException('InvalidSize')
            else:
                raise
        return True

    def showWorker(self, args):
        """
            Shows the details of an worker.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.wrk_id:
            workers = (self.api.read_worker(app_name, deployment_name, worker['wrk_id']) for worker in self.api.read_workers(app_name, deployment_name))
            print_worker_list(workers)
            return True
        else:
            try:
                worker = self.api.read_worker(
                    app_name,
                    deployment_name,
                    args.wrk_id)
            except GoneError:
                raise InputErrorException('WrongWorker')
            else:
                print_worker_details(worker)
                return True

    def removeWorker(self, args):
        """
            Removes an worker form a deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        try:
            self.api.delete_worker(app_name, deployment_name, args.wrk_id)
        except GoneError:
            raise InputErrorException('WrongWorker')
        return True

    def _restartWorker(self, app_name, deployment_name, wrk_id, command, params, size):
        self.api.delete_worker(app_name, deployment_name, wrk_id)
        self.api.create_worker(
            app_name,
            deployment_name,
            command,
            params,
            size)

    def _restartWorkers(self, app_name, deployment_name):
        workers = (self.api.read_worker(app_name, deployment_name, worker['wrk_id']) for worker in self.api.read_workers(app_name, deployment_name))
        for worker in workers:
            self._restartWorker(app_name, deployment_name, worker['wrk_id'], worker['command'], worker['params'], worker['size'])

    def restartWorker(self, args):
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'

        if args.wrk_id:
            try:
                worker = self.api.read_worker(app_name, deployment_name, args.wrk_id)
            except GoneError:
                raise InputErrorException('WrongWorker')

            self._restartWorker(app_name, deployment_name, worker['wrk_id'], worker['command'], worker['params'], worker['size'])
            return

        if args.all:
            self._restartWorkers(app_name, deployment_name)

    def addCron(self, args):
        """
            Adds the given worker to the deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.url:
            raise InputErrorException('NoCronURLGiven')
        self.api.create_cronjob(
            app_name,
            deployment_name,
            args.url)
        return True

    def showCron(self, args):
        """
            Shows the details of an worker.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.job_id:
            cronjobs = self.api.read_cronjobs(app_name, deployment_name)
            print_cronjob_list(cronjobs)
            return True
        else:
            try:
                cronjob = self.api.read_cronjob(
                    app_name,
                    deployment_name,
                    args.job_id)
            except GoneError:
                raise InputErrorException('NoSuchCronJob')
            else:
                print_cronjob_details(cronjob)
                return True

    def removeCron(self, args):
        """
            Removes an worker form a deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        try:
            self.api.delete_cronjob(app_name, deployment_name, args.job_id)
        except GoneError:
            raise InputErrorException('NoSuchCronJob')
        return True

    def _get_config_vars(self, app_name, deployment_name):
        try:
            addon = self.api.read_addon(app_name, deployment_name, CONFIG_ADDON)
            return addon['settings']['CONFIG_VARS']
        except (KeyError, GoneError):
            return {}

    def showConfig(self, args):
        """
            Shows the config variables
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'

        config = self._get_config_vars(app_name, deployment_name)

        if config:
            print_config(config, args.key)
        else:
            print '[ERROR] No configuration variables for this deployment.'

        return True

    def addConfig(self, args):
        """
            Adds the given variable and value to the config addon.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.variables:
            raise InputErrorException('NoVariablesGiven')

        args_variables, force = extract_flag_from_variables(args.variables,
                                                            ('-f', '--force'),
                                                            args.force_add)

        variables = parse_config_variables(args_variables, 'add')

        try:
            self.api.update_addon(
                app_name,
                deployment_name,
                CONFIG_ADDON,
                CONFIG_ADDON,
                settings=variables,
                force=force)
        except GoneError:
            # Add addon if it didn't exist.
            self.api.create_addon(app_name, deployment_name, CONFIG_ADDON, variables)
        except ThrottledError as te:
            # Overwrite the variable if didn't force, but type Yes
            question = raw_input('{} Do you really want to overwrite it? '
                                 'Type "Yes" without the quotes to proceed: '.format(te.message))
            if question.lower() == 'yes':
                self.api.update_addon(
                    app_name,
                    deployment_name,
                    CONFIG_ADDON,
                    CONFIG_ADDON,
                    settings=variables,
                    force=True)

        return True

    def removeConfig(self, args):
        """
            Remove the given variable if it exists.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.variables:
            raise InputErrorException('NoVariablesGiven')

        variables = parse_config_variables(args.variables, 'remove')

        try:
            self.api.update_addon(
                app_name,
                deployment_name,
                CONFIG_ADDON,
                CONFIG_ADDON,
                settings=variables)
        except GoneError:
            raise InputErrorException('WrongAddon')

        return True

    def listAddons(self, args):
        """
            Returns a list of all available addons
        """
        addons = self.api.read_addons()
        print_addons(addons)
        return True

    def addAddon(self, args):
        """
            Adds the given addon to the deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.addon:
            raise InputErrorException('NoAddonGiven')

        options = None
        if args.options:
            options = parse_additional_addon_options(args.options)

        try:
            self.api.create_addon(app_name, deployment_name, args.addon, options)
        except ConflictDuplicateError:
            raise InputErrorException('DuplicateAddon')
        except BadRequestError as e:
            if 'This is not a valid addon name' in str(e):
                raise InputErrorException('InvalidAddon')
            raise
        except ForbiddenError:
            raise InputErrorException('ForbiddenAddon')

        return True

    def showAddon(self, args):
        """
            Shows the details of an addon.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.addon:
            try:
                addons = self.api.read_addons(app_name, deployment_name)
            except:
                raise
            else:
                print_addon_list(addons)
                return True
        else:
            try:
                addon = self.api.read_addon(
                    app_name,
                    deployment_name,
                    args.addon)
            except GoneError:
                raise InputErrorException('WrongAddon')
            else:
                print_addon_details(addon)
                return True

    def showAddonCreds(self, args):
        """
            Print the creds.json of all Add-ons
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.addon:
            try:
                addons = self.api.read_addons(app_name, deployment_name)
            except:
                raise
            else:
                print_addon_creds(addons)
                return True
        else:
            try:
                addon = self.api.read_addon(
                    app_name,
                    deployment_name,
                    args.addon)
            except GoneError:
                raise InputErrorException('WrongAddon')
            else:
                print_addon_creds([addon])
                return True

    def updateAddon(self, args):
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        try:
            self.api.update_addon(
                app_name,
                deployment_name,
                args.addon_old,
                args.addon_new)
        except GoneError:
            raise InputErrorException('WrongAddon')
        else:
            return True

    def removeAddon(self, args):
        """
            Removes an addon form a deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        if not args.addon:
            raise InputErrorException('NoAddonGiven')
        try:
            self.api.delete_addon(app_name, deployment_name, args.addon)
        except GoneError:
            raise InputErrorException('WrongAddon')
        return True

    def showUser(self, args):
        """
            List users
        """

        app_name, deployment_name, obj = self._details(args.name)

        if deployment_name:
            print_user_list_deployment(obj)
        else:
            print_user_list_app(obj)

        return True

    def addUser(self, args):
        """
            Add a user specified by the e-mail address to an application or deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)

        if self.settings.prefix_project_name:
            if len(args.email.split(':')) != 2:
                prefix = self.api.read_users()[0]['email'].split(':')[0]
                args.email = '{}:{}'.format(prefix, args.email)

        try:
            if deployment_name:
                self.api.create_deployment_user(app_name, deployment_name, args.email, args.role)

            else:
                self.api.create_app_user(app_name, args.email, args.role)

        except ConflictDuplicateError:
            raise InputErrorException('UserBelongsToApp')
        except NotImplementedError:
            raise InputErrorException('CommandNotImplemented')
        return True

    def removeUser(self, args):
        """
            Remove a user specified by the user name or email address from an application or deployment.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if '@' in args.username:
            if deployment_name:
                users = self.api.read_deployment_users(app_name, deployment_name)
            else:
                users = self.api.read_app(app_name)['users']
            try:
                username = [user['username'] for user in users
                            if user['email'] == args.username][0]
            except IndexError:
                raise InputErrorException('RemoveUserGoneError')
        else:
            username = args.username
        try:
            if deployment_name:
                self.api.delete_deployment_user(app_name, deployment_name,
                                                username)

            else:
                self.api.delete_app_user(app_name, username)

        except GoneError:
            raise InputErrorException('RemoveUserGoneError')
        except NotImplementedError:
            raise InputErrorException('CommandNotImplemented')
        return True

    def log(self, args):
        """
        Show the log.
        """
        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        if not deployment_name:
            deployment_name = 'default'
        last_time = None
        while True:
            logEntries = []
            try:
                logEntries = self.api.read_log(
                    app_name,
                    deployment_name,
                    args.type,
                    last_time=last_time)
            except GoneError:
                raise InputErrorException('WrongApplication')
            if len(logEntries) > 0:
                last_time = datetime.fromtimestamp(float(logEntries[-1]["time"]))
                if args.type == 'worker' and args.wrk_id:
                    logEntries = filter(lambda entry:
                                        entry['wrk_id'] == args.wrk_id, logEntries)
                if args.filter:
                    if args.type in ["error", "worker"]:
                        logEntries = filter(
                            lambda entry: re.search(
                                re.compile(args.filter, re.IGNORECASE),
                                entry['message']),
                            logEntries)
                    if args.type == 'access':
                        logEntries = filter(lambda entry:
                                            re.search(
                                                re.compile(args.filter, re.IGNORECASE),
                                                entry['first_request_line'] +
                                                entry['referer'] +
                                                entry['user_agent'] +
                                                entry['remote_host']),
                                            logEntries)
                print_log_entries(logEntries, args.type)
            time.sleep(2)

    def log_from_now(self, app_name, deployment_name, type_):
        last_time = datetime.now() - timedelta(seconds=10)
        while True:
            logEntries = []
            try:
                logEntries = self.api.read_log(
                    app_name,
                    deployment_name,
                    type_,
                    last_time=last_time)
            except GoneError:
                raise InputErrorException('WrongApplication')

            if len(logEntries) > 0:
                last_time = datetime.fromtimestamp(float(logEntries[-1]["time"]))
                print_log_entries(logEntries, type_)

            time.sleep(1)

    def _get_or_create_deployment(self, app_name, deployment_name, clear_cache):
        try:
            if deployment_name == '':
                push_deployment_name = 'default'
            else:
                push_deployment_name = deployment_name

            deployment = self.api.read_deployment(
                app_name,
                push_deployment_name)

            if clear_cache:
                self._clear_cache(app_name, deployment_name, deployment)
        except GoneError:
            push_deployment_name = ''
            if deployment_name != '':
                push_deployment_name = deployment_name

            try:
                deployment = self.api.create_deployment(
                    app_name,
                    deployment_name=push_deployment_name)
            except GoneError:
                raise InputErrorException('WrongApplication')
            except ForbiddenError:
                raise InputErrorException('NotAllowed')

        return deployment, push_deployment_name

    def _push_cmd(self, deployment, push_deployment_name, source):
        cmd = None
        if deployment['branch'].startswith('bzr+ssh'):
            rcs = check_installed_rcs('bzr')
            if not rcs:
                raise InputErrorException('BazaarRequiredToPush')

            if source:
                cmd = [rcs, 'push', deployment['branch'], '-d', source]
            else:
                cmd = [rcs, 'push', deployment['branch']]
        elif deployment['branch'].startswith('ssh'):
            rcs = check_installed_rcs('git')
            if not rcs:
                raise InputErrorException('GitRequiredToPush')

            if push_deployment_name == 'default':
                git_branch = 'master'
            else:
                git_branch = push_deployment_name

            if source:
                git_dir = os.path.join(source, '.git')
                cmd = [
                    rcs,
                    '--git-dir=' + git_dir,
                    'push',
                    deployment['branch'],
                    git_branch]
            else:
                cmd = [rcs, 'push', deployment['branch'], git_branch]

        return cmd

    def push(self, args):
        """
            Push is actually only a shortcut for bzr and git push commands
            that automatically takes care of using the correct repository url.

            It queries the deployment details and uses whatever is in branch.

            If deployment exists will clear the buildpack cache if
            --clear-cache flag is set.

            If no deployment exists we automatically create one.


        """
        if not check_installed_rcs('bzr') and not check_installed_rcs('git'):
            raise InputErrorException('NeitherBazaarNorGitFound')

        if args.deploy and args.ship:
            raise InputErrorException('ShipAndDeploy')

        app_name, deployment_name = self.parse_app_deployment_name(args.name)
        deployment, push_deployment_name = self._get_or_create_deployment(app_name, deployment_name, args.clear_cache)

        cmd = self._push_cmd(deployment, push_deployment_name, args.source)

        try:
            check_call(cmd)
        except CalledProcessError, e:
            print str(e)
            sys.exit(1)

        if args.deploy:
            print
            print "Deploying newest version."
            self.redeploy(deployment)

        if args.ship:
            print
            print "Deploying newest version. Press Ctrl+C to open {} in your browser.".format(deployment['name'])
            self.redeploy(deployment)

            try:
                self.log_from_now(app_name, deployment_name, 'deploy')
            except KeyboardInterrupt:
                print
                print "Opening in browser..."
                self._open(self._get_deployment_url(deployment))

    def _clear_cache(self, app_name, deployment_name, deployment):
        subdomain = self._get_deployment_url(deployment).split('.', 1)[1]
        host_name = '{}@{}'.format(app_name, subdomain)

        cmd = ssh_cmd(host_name, 'delete-cache', deployment_name)

        try:
            check_call(cmd)
        except CalledProcessError:
            raise InputErrorException('ClearCacheFailed')

    def _get_public_key(self, app_name, application):
        host_name = self._get_repository_url(application)

        cmd = ssh_cmd(host_name, 'get-public-key')
        try:
            return check_output(cmd)
        except CalledProcessError:
            return "Not available. Have you added your Public Key?"

    def parse_app_deployment_name(self, name):
        match = re.match(
            '^([a-z][a-z0-9]*)/((?:[a-z0-9]+\.)*[a-z0-9]+)$',
            name)
        if match:
            app_name = match.group(1)
            deployment_name = match.group(2)
            return app_name, deployment_name

        match = re.match('^([a-z][a-z0-9]*)$', name)
        if match:
            app_name = match.group(1)
            deployment_name = ''
            return app_name, deployment_name

        raise ParseAppDeploymentName

    def does_app_exist(self, app_name):
        for app in self.api.read_apps():
            if app_name == app['name']:
                return True
        return False


class ParseAppDeploymentName(exceptions.Exception):
    """
        This Exception is raised if not a valid application name nor a
        valid application/deployment construct is given
    """
    pass



# -*- coding: utf-8 -*-

"""
    Copyright 2010 cloudControl UG (haftungsbeschraenkt)

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys
from os import path
import re
import json
from cctrl.oshelpers import recode_input
from cctrl.error import InputErrorException


def if_file_get_content(value):
    """
        if value is the path to a local file, read the content and return it
        otherwise just return value
    """
    file_path = path.abspath(value)
    if path.isfile(file_path):
        try:
            f = open(file_path, 'rU')
        except IOError:
            pass
        else:
            content = f.read()
            f.close()
            return content

    if sys.platform == 'win32':
        # in windows, sys.argv is encoded in windows-1252
        return value.decode('windows-1252').encode('UTF-8')
    else:
        # all others use the same encoding for stdin and argv
        return recode_input(value)


def parse_additional_addon_options(options):
    regex = re.compile('^(\w+)[=\s]?(.*)$')
    joined_options = ' '.join(options).split('--')
    result = {}
    for entry in joined_options:
        match = regex.match(entry)
        if match:
            if match.group(2):
                result[match.group(1)] = if_file_get_content(match.group(2).rstrip())
            else:
                result[match.group(1)] = 'true'
    return json.dumps(result)


def parse_config_variables(variables, method):
    if not variables:
        return None

    result = {}
    if method == 'remove':
        for var in variables:
            result[var.strip()] = None

    if method == 'add':
        for var in variables:
            if '=' in var:
                k, v = var.split('=', 1)
                result[k.strip()] = if_file_get_content(v.strip())
            else:
                result[var.strip()] = 'true'

    return json.dumps(result)


def extract_flag_from_variables(variables, flag_names, flag_value):
    flag_found = False

    for flag_name in flag_names:
        if flag_name in variables:
            variables.remove(flag_name)
            flag_found = True

    if flag_value and flag_found:
        raise InputErrorException('DuplicatedFlag')

    return variables, flag_value or flag_found



import unittest

from mock import patch, Mock

from cctrl.app import AppController
from cctrl.error import InputErrorException
from cctrl.settings import Settings


class AppControllerTestCase(unittest.TestCase):
    @patch('cctrl.app.parse_config_variables')
    def test_add_config(self, parse_config_variables):
        app = AppController(None, Settings())
        app.api = Mock()
        args = Mock()
        args.name = 'app/dep'
        args.force_add = False
        args.variables = ['foo=bar']
        app.addConfig(args)
        self.assertTrue(parse_config_variables.called)
        self.assertTrue(app.api.update_addon.called)

    @patch('cctrl.app.parse_config_variables')
    def test_add_config_force_flag(self, parse_config_variables):
        app = AppController(None, Settings())
        app.api = Mock()
        args = Mock()
        args.name = 'app/dep'
        args.force_add = True
        args.variables = ['foo=bar']
        app.addConfig(args)
        self.assertTrue(parse_config_variables.called)
        self.assertTrue(app.api.update_addon.called)

    @patch('cctrl.app.parse_config_variables')
    def test_add_config_force_arg(self, parse_config_variables):
        app = AppController(None, Settings())
        app.api = Mock()
        args = Mock()
        args.name = 'app/dep'
        args.force_add = False
        args.variables = ['foo=bar', '--force']
        app.addConfig(args)
        self.assertTrue(parse_config_variables.called)
        self.assertTrue(app.api.update_addon.called)

    @patch('cctrl.app.parse_config_variables')
    def test_add_config_no_variables_given(self, parse_config_variables):
        app = AppController(None, Settings())
        app.api = Mock()
        args = Mock()
        args.name = 'app/dep'
        args.force_add = False
        args.variables = None
        with self.assertRaises(InputErrorException) as nvg:
            app.addConfig(args)

        self.assertEqual(
            '[ERROR] You must provide some variables.',
            str(nvg.exception))
        self.assertFalse(parse_config_variables.called)
        self.assertFalse(app.api.update_addon.called)

    @patch('cctrl.app.parse_config_variables')
    def test_add_config_duplicated_flag(self, parse_config_variables):
        app = AppController(None, Settings())
        app.api = Mock()
        args = Mock()
        args.name = 'app/dep'
        args.force_add = True
        args.variables = ['foo=bar', '-f']
        with self.assertRaises(InputErrorException) as df:
            app.addConfig(args)

        self.assertEqual(
            '[ERROR] Please, specify a flag only once.',
            str(df.exception))
        self.assertFalse(parse_config_variables.called)
        self.assertFalse(app.api.update_addon.called)
